import collections
import datetime
import os
import random
import pandas as pd


def _ensure_dir(dirname):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)


def _make_list(x):
    if isinstance(x, int):
        return [x]

    if isinstance(x, str):
        return [x]

    if isinstance(x, collections.Sequence):
        return list(x)

    raise ValueError("Don't know how to make {} a list".format(x))


class QueryError(Exception):
    def __init__(self, msg, response):
        super(QueryError, self).__init__(msg)
        self.response = response


def iter_chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:(i + n)]


def random_dates(startdate, enddate, N, format="%Y-%m-%d"):
    # dates
    start = pd.to_datetime(startdate).to_pydatetime()
    end = pd.to_datetime(enddate).to_pydatetime()
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds

    def rand_time():
        return datetime.timedelta(seconds=random.randrange(int_delta))

    return [start + rand_time() for i in range(N)]



"""
This file is used to retrieve various datasets.

"""
import io
import pandas as pd
from .config import setup_logger
from .loader import load
from .bls import BLSData
from .socrata import SocrataData

LOGGER = setup_logger(__name__)


def _retrieve_test():
    df = pd.DataFrame({"A": [0, 1, 2],
                       "B": [3, 4, 5],
                       "C": [6, 7, 8]})

    return df, dict(index=[])


def _retrieve_state_fips():
    src = io.StringIO("""FIPS,Abbreviation,Name
    2,AK,Alaska
    28,MS,Mississippi
    1,AL,Alabama
    30,MT,Montana
    5,AR,Arkansas
    37,NC,North Carolina
    38,ND,North Dakota
    4,AZ,Arizona
    31,NE,Nebraska
    6,CA,California
    33,NH,New Hampshire
    8,CO,Colorado
    34,NJ,New jersey
    9,CT,Connecticut
    35,NM,New Mexico
    32,NV,Nevada
    10,DE,Delaware
    36,NY,New York
    12,FL,Florida
    39,OH,Ohio
    13,GA,Georgia
    40,OK,Oklahoma
    41,OR,Oregon
    15,HI,Hawaii
    42,PA,Pennsylvania
    19,IA,Iowa
    16,ID,Idaho
    44,RI,Rhode island
    17,IL,Illinois
    45,SC,South Carolina
    18,IN,Indiana
    46,SD,South Dakota
    20,KS,Kansas
    47,TN,Tennessee
    21,KY,Kentucky
    48,TX,Texas
    22,LA,Louisiana
    49,UT,Utah
    25,MA,Massachusetts
    51,VA,Virginia
    24,MD,Maryland
    23,ME,Maine
    50,VT,Vermont
    26,MI,Michigan
    53,WA,Washington
    27,MN,Minnesota
    55,WI,Wisconsin
    29,MO,Missouri
    54,WV,West Virginia
    56,WY,Wyoming
    """)
    return pd.read_csv(src), dict(index=[])


def _retrieve_state_employment():
    b = BLSData()

    states = load("state_fips")

    dfs = []
    for state_fips in states["FIPS"]:
        code = str(state_fips).zfill(2)
        codes = [
            "LASST{}0000000000003".format(code),
            "LASST{}0000000000006".format(code),
        ]
        LOGGER.debug("Querying bls for {}".format(codes))
        df = b.get(codes, startyear=2000, endyear=2017, nice_names=False)
        df["state"] = states.loc[states.FIPS == state_fips, "Name"].iloc[0]
        df.loc[df["variable"].str[-1] == "3", "variable"] = "UnemploymentRate"
        df.loc[df["variable"].str[-1] == "6", "variable"] = "LaborForce"
        df.set_index(["Date", "state", "variable"], inplace=True)
        df = df.unstack(level="variable")["value"]
        dfs.append(df)

    return pd.concat(dfs).sort_index(), dict(index=["Date", "state"])


def _retrieve_state_industry_employment():
    b = BLSData()

    states = load("state_fips")

    def get_codes(state_fips):
        code = str(state_fips).zfill(2)
        return {
            "SMS{}000000000000001".format(code): "total",
            "SMS{}000001000000001".format(code): "mining and logging",
            "SMS{}000002000000001".format(code): "construction",
            "SMS{}000003000000001".format(code): "manufacturing",
            "SMS{}000004000000001".format(code): "trade, transportation and utilities",
            "SMS{}000005000000001".format(code): "information",
            "SMS{}000005500000001".format(code): "financial activities",
            "SMS{}000006000000001".format(code): "professional and business services",
            "SMS{}000006561000001".format(code): "education",
            "SMS{}000006562000001".format(code): "healthcare",
            "SMS{}000007000000001".format(code): "leisure and hospitality",
            "SMS{}000008000000001".format(code): "other services",
            "SMS{}000009000000001".format(code): "government"
        }

    dfs = []
    for state_fips in states["FIPS"]:
        codes = get_codes(state_fips)
        LOGGER.debug("Querying bls for {}".format(codes))
        df = b.get(
            list(codes.keys()), startyear=2000, endyear=2017,
            nice_names=False
        )
        df.replace({"variable": codes}, inplace=True)
        df["state"] = states.loc[states.FIPS == state_fips, "Name"].iloc[0]
        df.set_index(["Date", "state", "variable"], inplace=True)
        df = df.unstack(level="variable")["value"]
        dfs.append(df)

    return pd.concat(dfs).sort_index(), dict(index=["Date", "state"])


def _retrieve_goodreads_books():
    LOGGER.debug("Downloading goodreads books.csv from github")
    url = "https://raw.githubusercontent.com/zygmuntz/goodbooks-10k/"
    url += "c8a6e0a9a3b620c3f89301b0b3dc2a6653972294/books.csv"
    return pd.read_csv(url), dict(index=[])


def _retrieve_goodreads_ratings():
    LOGGER.debug("Downloading goodreads ratings.csv from github")
    url = "https://raw.githubusercontent.com/zygmuntz/goodbooks-10k/"
    url += "c8a6e0a9a3b620c3f89301b0b3dc2a6653972294/ratings.csv"
    return pd.read_csv(url), dict(index=[])


def _retrieve_goodreads_tags():
    LOGGER.debug("Downloading goodreads tags.csv from github")
    url = "https://raw.githubusercontent.com/zygmuntz/goodbooks-10k/"
    url += "c8a6e0a9a3b620c3f89301b0b3dc2a6653972294/tags.csv"
    return pd.read_csv(url), dict(index=[])


def _retrieve_goodreads_book_tags():
    LOGGER.debug("Downloading goodreads book_tags.csv from github")
    url = "https://raw.githubusercontent.com/zygmuntz/goodbooks-10k/"
    url += "c8a6e0a9a3b620c3f89301b0b3dc2a6653972294/book_tags.csv"
    return pd.read_csv(url), dict(index=[])


def _get_airline_data(url):
    LOGGER.debug("Downloading airline data from {}".format(url))
    df = pd.read_csv(url)
    df["Date"] = pd.to_datetime(df["FlightDate"])
    df.drop("FlightDate", axis=1, inplace=True)
    bad_cols = list(filter(lambda x: x.startswith("Unnamed"), list(df)))
    df.drop(bad_cols, axis=1, inplace=True)

    for col in ["CRSDepTime", "CRSArrTime"]:
        LOGGER.debug("Converting column {} to datetime".format(col))
        dt_string = df["Date"].astype(str) + df[col].astype(str).str.zfill(4)
        df[col] = pd.to_datetime(dt_string, format="%Y-%m-%d%H%M")

    for col in ["DepTime", "ArrTime"]:
        LOGGER.debug("Converting column {} to datetime".format(col))
        t_string = df[col].astype(str).str[:-2].str.zfill(4)
        dt_string = df["Date"].astype(str) + t_string
        df[col] = pd.to_datetime(dt_string, format="%Y-%m-%d%H%M",
                                 errors="coerce")

    # If the delay value is a NaN then no delay for any of these reasons...
    # Replace with 0.0
    delays = [
        "WeatherDelay", "CarrierDelay", "NASDelay", "SecurityDelay",
        "LateAircraftDelay"
    ]
    df.loc[:, delays] = df.loc[:, delays].fillna(0.0)

    return df, dict(index=[])


def _retrieve_airline_performance_dec16():
    url = "https://storage.googleapis.com/qeds/data/"
    url += "December2016_ontimeflights.csv.zip"
    return _get_airline_data(url)


def _retrieve_airline_performance_nov16():
    url = "https://storage.googleapis.com/qeds/data/"
    url += "November2016_ontimeflights.csv.zip"
    return _get_airline_data(url)


def _retrieve_airline_carrier_codes():
    url = "https://storage.googleapis.com/qeds/data/Carrier_Codes.csv"
    return pd.read_csv(url).set_index("Code"), dict(index=["Code"])


def _retrieve_nyc_employee():
    # Download data
    sd = SocrataData("4qxi-jgbe", "NYCOpenData")
    df = sd.get(limit=None)  # Get all observations

    #
    # Clean data up
    #

    # Convert columns to numeric
    num_columns = ["base_salary", "fiscal_year", "ot_hours",
                   "regular_gross_paid", "regular_hours", "total_ot_paid",
                   "total_other_pay"]
    for col in num_columns:
        df[col] = pd.to_numeric(df[col])

    # Convert to datetime
    str_f = "%Y-%m-%dT00:00:00.000"
    df.replace("9999-12-31T00:00:00.000", pd.np.nan, inplace=True)
    df["agency_start_date"] = pd.to_datetime(df["agency_start_date"],
                                             format=str_f)

    # Stupid strings
    str_columns = ["agency_name", "first_name", "mid_init", "last_name",
                   "leave_status_as_of_july_31", "pay_basis",
                   "title_description", "work_location_borough"]
    for col in str_columns:
        df[col] = (df[col].str.strip()  # Drop stupid spaces
                          .str.upper()  # Capitalize
                          .replace(pd.np.nan, ""))

    return df, dict(index=[])


def _retrieve_chipotle_raw():
    url = "https://raw.githubusercontent.com/TheUpshot/"
    url += "chipotle/master/orders.tsv"
    return pd.read_csv(url, sep="\t"), dict(index=[])



import string
import warnings
from ..config import _get_option


def validate_api_key(key):
    API_KEY_LENGTH = 32
    if len(key) > API_KEY_LENGTH:
        key = key[:API_KEY_LENGTH]
        msg = "API key too long, using first {} characters".format(API_KEY_LENGTH)
        warnings.warn(msg)
    elif len(key) < API_KEY_LENGTH:
        msg = "API key {} too short. Should be {} chars".formaT(key, API_KEY_LENGTH)
        raise ValueError(msg)

    if not all(i in string.hexdigits for i in key):
        msg = "API key {} contains invalid characters".format(key)
        raise ValueError(msg)


_get_option("bls", "api_key").validator = validate_api_key


LIMITS = {
    "years_per_query": 20,
    "series_per_query": 50,
}

# see https://www.bls.gov/developers/api_faqs.htm
BLS_STATUS_CODE_REASONS = {
    400: "Your request did not follow the correct syntax.",
    401: "You are not authorized to make this request.",
    404: "Your request was not found and/or does not exist.",
    429: "You have made too many requests.",
    500: ("The server has encountered an unexpected condition, and the " +
          "request cannot be completed.")
}



import json
import os
import textwrap

import pandas as pd
import requests
from requests.adapters import HTTPAdapter
from ..config import options
from ..util import _make_list, QueryError, _ensure_dir

_ensure_dir(options["uscensus.data_dir"])


def _update_data_file():
    url = "https://api.census.gov/data.json"
    r = requests.get(url)
    file_path = os.path.join(options["uscensus.data_dir"], "data.json")
    with open(file_path, "w") as f:
        f.write(json.dumps(r.json()))


def _load_metadata():
    data_fn = os.path.join(options["uscensus.data_dir"], "data.json")
    if not os.path.isfile(data_fn):
        _update_data_file()

    with open(data_fn, "r") as f:
        _DATA_RAW = json.load(f)

    _DATA = pd.DataFrame(_DATA_RAW["dataset"])
    _DATA["c_dataset"] = _DATA["c_dataset"].apply(lambda x: "/".join(x))
    return _DATA_RAW, _DATA


_DATA_RAW, _DATA = _load_metadata()


def query_predicate_string(name, arg):
    arg = _make_list(arg)
    if len(arg) > 0:
        out = "&{}=".format(name)
        out += "&{}=".format(name).join(str(i) for i in arg)
        return out
    else:
        return ""

    raise ValueError("Don't know how to handle query predicate arg {}".format(arg))


def geo_predicate_string(name, arg):
    arg = _make_list(arg)
    return name + ":" + ",".join(str(i) for i in arg)


class CensusData(object):
    def __init__(self, url=None, key=None):
        """
        Parameters
        ----------
        url : string, optional(default={url})
            The API url to be used when requesting data from the BLS

        key : string, optional
            The API registrationKey to be used when requesting data.
            There are three possibilities for determining which api key is
            used:

            1. User argument: if the argument supplied here is a valid
               BLS registration key, then that value will be used
            2. Environment variable: if step 1 fails, we will attempt to use
               the value of the environment variable {key_env_name}.
            3. qeds conf: if that fails, we will attempt to look up the
               value in section `bls.api_key` if your qeds conf file

            If all three of those fail, we throw an error whose message
            contains instructions for how to obtain a registration key.

            If either step 2 or step 2 succeeds, we will store the api key
            in the qedsm conf fil under `bls.api_key`. This means you
            should only need to supply a key once per machine.
        """.format(
            url=options["uscensus.api_url"],
            key_env_name=options["uscensus.environment_variable"]
        )
        update_config = True
        if key is None:
            KEY_ENV_NAME = options["uscensus.environment_variable"]
            if KEY_ENV_NAME in os.environ:
                key = os.environ[KEY_ENV_NAME]
            elif options["uscensus.api_key"] is not None:
                key = options["uscensus.api_key"]
                update_config = False
            else:
                raise EnvironmentError("Census API key not detected")

        if update_config:
            options["uscensus.api_key"] = key

        if url is None:
            url = options["uscensus.api_url"]

        self.key = key
        self.url = url

        self.sess = requests.Session()
        self.sess.mount(self.url, HTTPAdapter(max_retries=3))

        # NOTE: subclasses must define self.dataset, self.meta, and
        # self.vars_df

    def _geography_query(self, kwargs):
        """
        Given the information about the state and county, add `for` and `in`
        arguments to params
        """
        state = _make_list(kwargs.pop("state", []))
        county = _make_list(kwargs.pop("county", []))
        msa = _make_list(kwargs.pop("msa", []))

        out = ""
        if len(msa) > 0:
            out += "&for="
            out += geo_predicate_string(
                "metropolitan statistical area/micropolitan statistical area",
                msa
            )
            return out

        if len(county) == 0:
            if len(state) == 0:
                m = "Both state and county were empty. "
                m += "Some geography must be given"
                raise ValueError(m)

            # state is not empty
            out += "&for=" + geo_predicate_string("state", state)
            return out

        # county is not empty
        out += "&for=" + geo_predicate_string("county", county)

        if len(state) > 0:
            # also have in clause for states
            out += "&in=" + geo_predicate_string("state", state)

        return out

    def get(self, variables, start_time=None, end_time=None, timeout=None,
            **kwargs):
        """
        Get the specified variables from the data set

        Parameters
        ----------
        variables: list(str)
            A list of variables to obtain

        start_time, end_time: str, optional(default=None)
            A YYYY-MM-DD string specifying the starting and ending time. Only
            applicable for time series datasets

        **kwargs
            All other keyword arguments are used as predicates in the query
        """
        variables = _make_list(variables)

        self.validate_vars(variables)
        var_string = ",".join(variables)
        query = "?get={}&key={}".format(var_string, self.key)

        # all kwargs must also be valid varirables, except for state, us,
        # and county, which are handled separately
        if "us" in kwargs:
            # if US is present, then we will skip the state/county args
            query += "&for=us:*"
            kwargs.pop("us")
        else:
            query += self._geography_query(kwargs)

        self.validate_vars(kwargs.keys())

        for (k, v) in kwargs.items():
            if k != "state" and k != "county":
                query += query_predicate_string(k, v)

        r = self.sess.get(self.url + self.dataset + query, timeout=timeout)
        if r.status_code != 200:
            msg = "Query failed with status code {}. ".format(r.status_code)
            msg += "Response from server was\n{}".format(r.content)
            raise QueryError(msg, r)

        try:
            js = r.json()
            df = pd.DataFrame(js[1:], columns=js[0])
        except:  # NOQA
            if "SIC" in kwargs or "SIC" in variables:
                try:
                    df = pd.read_json(r.content.replace(b"\\", b""))
                    df.columns = df.iloc[0, :].tolist()
                    df.drop(0, axis=0, inplace=True)
                except:  # NOQA
                    msg = "Couldn't parse query result into DataFrame."
                    raise QueryError(msg, r)
            else:
                msg = "Query failed with status code {}. ".format(r.status_code)
                msg += "Response from server was\n{}".format(r.content)
                raise QueryError(msg, r)

        key = ("metropolitan statistical area/"
               "micropolitan statistical area")
        df.rename(
            columns={key: "MSA"},
            inplace=True
        )

        for k in df.columns:
            if k in ["state", "county", "us", "MSA", "zipcode"]:
                df[k] = df[k].astype(int)
                continue

            if k == "SIC":
                try:
                    df[k] = df[k].astype(int)
                except ValueError:
                    continue

            dtype_str = self.vars_df.loc["predicateType", k]
            if dtype_str == "string":
                # nothing to do
                continue

            if isinstance(dtype_str, float):
                # NaN -- nothing to do
                continue

            try:
                df[k] = df[k].astype(dtype_str)
            except TypeError:
                # couldn't do conversion -- probably a None somewhere
                # we will skip it for now and let the user worry about it
                continue

        return df

    def _variables_file_name(self):
        name = self.dataset.replace("/", "_")
        return os.path.join(options["uscensus.data_dir"], "{}.json".format(name))

    def _get_variables_file(self):
        fn = self._variables_file_name()
        if os.path.isfile(fn):
            with open(fn, "r") as f:
                raw = json.load(f)
        else:
            r = self.sess.get(self.meta["c_variablesLink"].iloc[0])
            raw = r.json()
            with open(fn, "w") as f:
                f.write(json.dumps(r.json()))

        df = pd.DataFrame(raw["variables"])
        if "required" in df.columns:
            df.loc["required"] = df.loc["required"].fillna("False")
        return df

    # Validation
    def validate_vars(self, variables):
        for var in variables:
            if var.upper() not in self.vars_df.columns:
                varstring = textwrap.wrap(", ".join(self.vars_df.columns))
                m = "\nInvalid variable {} requested. ".format(var)
                m += "Possilble choices are:\n"
                m += textwrap.indent("\n".join(varstring), "    ")
                raise ValueError(m)


class CountyBusinessPatterns(CensusData):

    def __init__(self, year, url=None, key=None):
        super(CountyBusinessPatterns, self).__init__(url, key)
        self.year = year
        self.dataset = "{}/cbp".format(year)

        meta = _DATA[
            (_DATA["c_dataset"] == "cbp") &
            (_DATA["temporal"] == "{year}/{year}".format(year=self.year))
        ]

        if meta.shape[0] == 0:
            years = (
                _DATA[_DATA["c_dataset"] == "cbp"]
                ["temporal"]
                .str.split("/")
                .str.get(0)
                .astype(int)
            )
            min_year = years.min()
            max_year = years.max()
            m = "Invalid year {}. Must be in range {}-{}".format(year, min_year, max_year)
            raise ValueError(m)

        self.meta = meta
        self.vars_df = self._get_variables_file()


class ZipBusinessPatterns(CensusData):

    def __init__(self, year, url=None, key=None):
        super(ZipBusinessPatterns, self).__init__(url, key)
        self.year = year
        self.dataset = "{}/zbp".format(year)

        meta = _DATA[
            (_DATA["c_dataset"] == "zbp") &
            (_DATA["temporal"] == "{year}/{year}".format(year=self.year))
            ]

        if meta.shape[0] == 0:
            years = (
                _DATA[_DATA["c_dataset"] == "zbp"]
                ["temporal"]
                .str.split("/")
                .str.get(0)
                .astype(int)
            )
            min_year = years.min()
            max_year = years.max()
            m = "Invalid year {}. Must be in range {}-{}".format(year, min_year, max_year)
            raise ValueError(m)

        self.meta = meta
        self.vars_df = self._get_variables_file()

    def _geography_query(self, kwargs):
        """
        Extract zipcode and state from kwargs and build appropriate geography
        query
        """
        out = ""
        zipcode = _make_list(kwargs.pop("zipcode", []))
        state = _make_list(kwargs.pop("state", []))

        if len(zipcode) == 0:
            if len(state) == 0:
                m = "Both state and zipcode were empty. "
                m += "Some geography must be given"
                raise ValueError(m)

            # state is not empty
            out += "&for=" + geo_predicate_string("zipcode", "*")
            out += "&in=" + geo_predicate_string("state", state)
            return out

        # zipcode is not empty
        out += "&for=" + geo_predicate_string("zipcode", zipcode)

        if len(state) > 0:
            # also have in clause for states
            out += "&in=" + geo_predicate_string("state", state)

        return out


if __name__ == '__main__':
    cbp = CountyBusinessPatterns(2010)



import datetime
import os

import pandas as pd
import requests
from requests.adapters import HTTPAdapter

from ..config import options, setup_logger
from ..util import _make_list, QueryError, iter_chunks

from .util import LIMITS, BLS_STATUS_CODE_REASONS

LOGGER = setup_logger(__name__)


class BLSData(object):
    def __init__(self, url=None, key=None):
        """
        Parameters
        ----------
        url : string, optional(default={url})
            The API url to be used when requesting data from the BLS

        key : string, optional
            The API registrationKey to be used when requesting data.
            There are three possibilities for determining which api key is
            used:

            1. User argument: if the argument supplied here is a valid
               BLS registration key, then that value will be used
            2. Environment variable: if step 1 fails, we will attempt to use
               the value of the environment variable {key_env_name}.
            3. qeds conf: if that fails, we will attempt to look up the
               value in section `bls.api_key` if your qeds conf file

            If all three of those fail, we throw an error whose message
            contains instructions for how to obtain a registration key.

            If either step 2 or step 2 succeeds, we will store the api key
            in the qedsm conf fil under `bls.api_key`. This means you
            should only need to supply a key once per machine.
        """.format(
            url=options["bls.api_url"],
            key_env_name=options["bls.environment_variable"]
        )

        update_config = True
        if key is None:
            KEY_ENV_NAME = options["bls.environment_variable"]
            if KEY_ENV_NAME in os.environ:
                key = os.environ[KEY_ENV_NAME]
            elif options["bls.api_key"] is not None:
                key = options["bls.api_key"]
                update_config = False
            else:
                url = "https://data.bls.gov/registrationEngine/"
                msg = "BLS API key not detected. Please make one at {}".format(url)
                msg += " and call `qeds.options['bls.api_key']=key`"
                raise EnvironmentError(msg)

        if update_config:
            options["bls.api_key"] = key

        if url is None:
            url = options["bls.api_url"]

        self.key = key
        self.url = url
        self.headers = {"Content-Type": "application/json"}

        self.sess = requests.Session()
        self.sess.mount(self.url, HTTPAdapter(max_retries=3))

    def get(self, series, startyear=None, endyear=None, nice_names=True,
            wide=False):
        """
        Get requested ``series`` from ``startyear`` to ``endyear`` as a
        pandas DataFrame

        Parameters
        ----------
        series : string or list(string)
            A valid BLS series name, or list of series names

        startyear: int, optional(default=endyear - 19)
            The starting year for which to obtain data. The default is 19
            years before the endyear, so 20 years of data are obtained

        endyear: int, optional(default=current_year)
            The ending year for data lookup. The default is the current year

        nice_names: bool, optional(default=True)
            Fill the `variable` column of the output with the series title
            instead of the series id. For example, if you request series
            ``LASST040000000000006``, and pass ``nice_names=True``, then the
            corresponding entry in the ``variable`` column will be
            ``Labor Force: Arizona (S)`` instead of the series id.

        wide: bool, optional(default=False)
            Toggles the return of a wide DataFrame with the index being the
            date and one variable per column

        Returns
        -------
        df : pandas.DataFrame
            A pandas DataFrame contianing the requested series

        """

        series = _make_list(series)
        nyear = LIMITS["years_per_query"]
        nseries = LIMITS["series_per_query"]

        if endyear is None:
            endyear = datetime.datetime.now().year
        if startyear is None:
            startyear = endyear - (nyear - 1)

        # Chunk on years if user asked for more than 20 years
        if endyear - startyear > (nyear - 1):
            to_cat = []
            years = range(startyear, endyear + 1)
            for year_chunks in iter_chunks(years, nyear):
                df = self.get(
                    series, year_chunks[0], year_chunks[-1], nice_names, wide
                )
                to_cat.append(df)

            return pd.concat(to_cat, ignore_index=True)

        # Chunk on series if user asked for more than 50 series
        if len(series) > nseries:
            to_cat = []
            for series_chunk in iter_chunks(series, nseries):
                df = self.get(
                    series_chunk, startyear, endyear, nice_names, wide
                )
                to_cat.append(df)

            return pd.concat(to_cat, ignore_index=True)

        body = {
            "startyear": startyear,
            "endyear": endyear,
            "seriesid": series,
            "registrationKey": self.key,
            "catalog": True if nice_names else False,
        }
        res = self.sess.post(self.url, json=body)

        if res.status_code == 200:
            data = res.json()
        elif res.status_code == 202:
            # 202 means the bls servers are still processing...
            # we'll try to read it anyway, even though there will probably
            # be errors later
            data = res.json()
        elif res.status_code in BLS_STATUS_CODE_REASONS:
            msg = "Request failed with code {} and message ".format(res.status_code)
            msg += BLS_STATUS_CODE_REASONS(res.status_code)
            raise QueryError(msg, res)
        else:
            msg = "Request failed unexpectedly with code {}".format(res.status_code)
            raise QueryError(msg, res)

        if data.get("status", "") == "REQUEST_NOT_PROCESSED":
            msg = "Request was not processed"
            if len(data.get("message", "")) > 0:
                msg += ". Received message {}".format(data["message"])

            raise QueryError(msg, res)


        # at this point we should have data

        dfs = []
        for series in data["Results"]["series"]:
            df = pd.DataFrame(
                series["data"],
                columns=["value", "year", "period"]
            )
            if df.shape[0] == 0:
                LOGGER.debug("Query was empty for " + series['seriesID'])
                continue

            # see if quarterly, monthly, or annual
            freq = series["data"][0]["period"][0]
            if freq == "M":
                df["Date"] = pd.to_datetime(
                    df["year"] + df["period"], format="%YM%m"
                )
            elif freq == "A":
                df["Date"] = pd.to_datetime(df["year"], format="%Y")
            elif freq == "Q":
                df["Date"] = pd.to_datetime(df["year"] + df["period"].str[2])
            else:
                msg = "Unknown frequency {}. Please open an issue".format(freq)
                raise ValueError(msg)

            df.drop(["year", "period"], axis=1, inplace=True)

            if nice_names:
                if "series_title" in series.get("catalog", []):
                    df["variable"] = series["catalog"]["series_title"]
            else:
                df["variable"] = series["seriesID"]

            try:
                df["value"] = df["value"].astype(float)
            except ValueError:
                pass

            dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)

        if wide:
            return df.set_index(["Date", "variable"]).unstack()["value"]
        else:
            return df



import os
import string
import warnings
import pandas as pd
from ..config import options, _get_option


def validate_api_key(key):
    API_KEY_LENGTH = 40
    if len(key) > API_KEY_LENGTH:
        key = key[:API_KEY_LENGTH]
        msg = "API key too long, using first {} characters".format(API_KEY_LENGTH)
        warnings.warn(msg)
    elif len(key) < API_KEY_LENGTH:
        msg = "API key {} too short. Should be {} chars".format(key, API_KEY_LENGTH)
        raise ValueError(msg)

    if not all(i in string.hexdigits for i in key):
        msg = "API key {} contains invalid characters".format(key)
        raise ValueError(msg)


_get_option("uscensus", "api_key").validator = validate_api_key

cbp_industry_var = {
    1986: "SIC",
    1987: "SIC",
    1988: "SIC",
    1989: "SIC",
    1990: "SIC",
    1991: "SIC",
    1992: "SIC",
    1993: "SIC",
    1994: "SIC",
    1995: "SIC",
    1996: "SIC",
    1997: "SIC",
    1998: "NAICS1997",
    1999: "NAICS1997",
    2000: "NAICS1997",
    2001: "NAICS1997",
    2002: "NAICS1997",
    2003: "NAICS2002",
    2004: "NAICS2002",
    2005: "NAICS2002",
    2006: "NAICS2002",
    2007: "NAICS2002",
    2008: "NAICS2007",
    2009: "NAICS2007",
    2010: "NAICS2007",
    2011: "NAICS2007",
    2012: "NAICS2012",
    2013: "NAICS2012",
    2014: "NAICS2012",
    2015: "NAICS2012"
}


def update_fips_2010():
    url = "https://www2.census.gov/geo/docs/reference"
    url += "/codes/files/national_county.txt"
    df = pd.read_csv(url, header=None)
    df.columns = [
        "State_Name",
        "State",
        "County",
        "County_Name",
        "Class_FIPS"
    ]
    df.set_index(["State", "County"], inplace=True)
    df.index.name = "FIPS"
    df.sort_index(inplace=True)
    fn = os.path.join(options["uscensus.data_dir"], "fips2010.csv")
    df.to_csv(fn)
    return df


def get_fips_2010():
    path = os.path.join(options["uscensus.data_dir"], "fips2010.csv")
    if os.path.isfile(path):
        return pd.read_csv(path, index_col=[0, 1])

    else:
        return update_fips_2010()


def get_naics2002_to_sics():
    path = os.path.join(options["uscensus.data_dir"], "naics2002_to_sic.csv")
    if os.path.isfile(path):
        return pd.read_csv(path, index_col=0)
    else:
        return update_naics2002_to_sics()


# stuff to make working with this data bearable
def update_naics2002_to_sics():
    url = "https://www.census.gov/eos/www/naics/"
    url += "concordances/1987_SIC_to_2002_NAICS.xls"
    df = pd.read_excel(url, skip_footer=1)
    fn = os.path.join(options["uscensus.data_dir"], "naics2002_to_sic.csv")
    df.to_csv(fn)
    return df


def naics2002_to_sics(start):
    df = get_naics2002_to_sics()
    subsets = df[df["2002 NAICS"].astype(str).str.startswith(start)]
    return subsets["SIC"].dropna().astype(int).unique()


def update_naics_crosswalk():
    url = "https://www.census.gov/eos/www/naics/concordances"
    url += "/{}_to_{}_NAICS.xlsx"
    df = pd.read_excel(
        url.format(2017, 2012),
        skiprows=2,
        usecols=[0, 2],
        names=["NAICS2017", "NAICS2012"],
    ).merge(pd.read_excel(
        url.format(2012, 2007)[:-1],
        skiprows=2,
        usecols=[0, 2],
        names=["NAICS2012", "NAICS2007"],
    )).merge(pd.read_excel(
        url.format(2007, 2002)[:-1],
        skiprows=2,
        usecols=[0, 2],
        names=["NAICS2007", "NAICS2002"],
    ))
    fn = os.path.join(options["uscensus.data_dir"], "naics_crosswalk.csv")
    df.to_csv(fn)
    return df


def get_naics_crosswalk():
    path = os.path.join(options["uscensus.data_dir"], "naics_crosswalk.csv")
    if os.path.isfile(path):
        return pd.read_csv(path, index_col=0)
    else:
        return update_naics_crosswalk()


# SIC codes
def update_sic86():
    url = "http://www2.census.gov/programs-surveys/cbp/technical-documentation"
    url += "/records-layouts/sic-code-descriptions/sic86.txt"
    df = pd.read_csv(url, sep="    ")
    df.columns = ["SIC", "NAME"]
    df.to_csv(os.path.join(options["uscensus.data_dir"], "sic86.csv"))
    return df


def get_sic86():
    path = os.path.join(options["uscensus.data_dir"], "sic86.csv")
    if os.path.isfile(path):
        return pd.read_csv(path, index_col=0)
    else:
        return update_sic86()


def update_sic87():
    url = "http://www2.census.gov/programs-surveys/cbp/technical-documentation"
    url += "/records-layouts/sic-code-descriptions/sic87.txt"
    df = pd.read_csv(url, sep="    ")
    df.columns = ["SIC", "NAME"]
    df.to_csv(os.path.join(options["uscensus.data_dir"], "sic87.csv"))
    return df


def get_sic87():
    path = os.path.join(options["uscensus.data_dir"], "sic87.csv")
    if os.path.isfile(path):
        return pd.read_csv(path, index_col=0)
    else:
        return update_sic87()



"""
This file reads in a configuration file if one already exists and
if one does not exist then it creates one
"""
import configparser
import os
import pathlib
import logging
import warnings

# Get home directory and config file
_home = str(pathlib.Path.home())
_BASE_PATH = os.path.join(_home, ".qeds")
CFG_FILE = os.path.join(_BASE_PATH, "config.ini")
BASE_DATA_DIR = os.path.join(_BASE_PATH, "data")

if not os.path.isdir(_BASE_PATH):
    os.mkdir(_BASE_PATH)


class Option:
    __slots__ = ["name", "default", "doc", "validator"]

    def __init__(self, name, default, doc, validator):
        self.name = name
        self.default = default
        self.doc = doc
        self.validator = validator

    def __str__(self):
        info = dict(name=self.name, default=self.default, doc=self.doc)
        if self.default is not None:
            msg = "  - {name} (default={default}) : {doc}"
        else:
            msg = "  - {name} : {doc}"

        return msg.format(**info)

    def __repr__(self):
        return self.__str__()


def _no_validation(x):
    pass


def _member_validation(allowed):
    def func(val):
        if val not in allowed:
            msg = "Value {} not allowed. Acceptable values are {}"
            raise configparser.Error(msg)

    return func


# dict holding all config options. Maps from vconf section name to a list of
# options
_valid_options = {
    "PATHS": [
        Option(
            "base",
            str(_BASE_PATH),
            "Default directory for all things qeds",
            _no_validation
        ),
        Option(
            "data",
            str(BASE_DATA_DIR),
            "Default directory for saving loaded data",
            _no_validation
        ),
    ],
    "options": [
        Option(
            "file_format",
            "csv",
            "File format for saving loaded data",
            _member_validation(["pkl", "csv", "feather"])
        ),
        Option(
            "log_level",
            "CRITICAL",
            "Default level for filtering logging messages",
            _member_validation([
                "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"
            ])
        )
    ],
    "uscensus": [
        Option(
            "api_key",
            None,
            """API key for accessing data from US Census Bureau\
            API. Obtain one at https://api.census.gov/data/key_signup.html""",
            _no_validation
        ),
        Option(
            "data_dir",
            os.path.join(_BASE_PATH, "uscensus", "data"),
            "Directory to store the supplementary files for uscensus api",
            _no_validation
        ),
        Option(
            "environment_variable",
            "USCENSUS_API_KEY",
            "Environment variable to search for api_key for US Census Bureau",
            _no_validation
        ),
        Option(
            "api_url",
            "https://api.census.gov/data/",
            "URL through which to access the US Census Bureau API",
            _no_validation
        )
    ],
    "bls": [
        Option(
            "api_key",
            None,
            """API key for accessing data from US Bureau of Labor Statistics\
            API. Obtain one at https://data.bls.gov/registrationEngine/""",
            _no_validation
        ),
        Option(
            "environment_variable",
            "BLS_API_KEY",
            "Environment variable to search for api_key for BLS",
            _no_validation
        ),
        Option(
            "api_url",
            "https://api.bls.gov/publicAPI/v2/timeseries/data/",
            "URL through which to access the BLS API",
            _no_validation
        )
    ],
    "socrata": [
        Option(
            "api_key",
            None,
            """API key for accessing data from the Socrata API. Obtain one at\
            https://dev.socrata.com/register/""",
            _no_validation
        ),
        Option(
            "environment_variable",
            "SOCRATA_API_KEY",
            "Environment variable to search for api_key for Socrata",
            _no_validation
        ),
    ]
}


def _get_option(section, name, warn=False):
    if section not in _valid_options:
        msg = "Unknown config section {}. Valid sections are {}"
        valid = list(_valid_options.keys())
        m = msg.format(section, valid)
        if warn:
            warnings.warn(m)
            return
        raise ValueError(m)

    options = _valid_options[section]

    if name not in [o.name for o in options]:
        msg = "Unknown option {}. Known options are {}"
        valid = [o.name for o in options]
        m = msg.format(section, valid)
        if warn:
            warnings.warn(m)
            return
        raise ValueError(m)

    _opt = None
    for o in options:
        if o.name == name:
            _opt = o
            break

    if _opt is None:
        m = "Not sure how I got here... file a bug report."
        if warn:
            warnings.warn(m)
            return
        raise ValueError(m)

    return _opt


def _validate_config_setting(section, name, value):
    option = _get_option(section, name)
    option.validator(value)


# -------------- #
# Convenient API #
# -------------- #

class _DictOptions(object):
    def __init__(self):
        self.vconf = configparser.ConfigParser()
        self.load_config()

    def load_config(self):
        if os.path.exists(CFG_FILE):
            self.vconf.read(CFG_FILE)

        # add defaults to config
        for (sec, opts) in _valid_options.items():
            if not self.vconf.has_section(sec):
                self.vconf.add_section(sec)

            for o in opts:
                if self.vconf.has_option(sec, o.name):
                    continue
                # vconf doesn't have option, write one if we have a default
                if o.default is not None:
                    self.vconf.set(sec, o.name, o.default)

        # save updated config
        self.write_config()

    def validate_config(self, warn=True):
        """
        Validate all settings in the config

        Parameters
        ----------
        warn : bool, optional(default=True)
            Indicator for whether the function should warn or raise errors
            when an invalid config value is found. Default is warn
        """
        for (section, options) in self.vconf.items():
            if section == "DEFAULT":
                continue
            if section not in _valid_options:
                warnings.warn("Unknown section {}".format(section))

            for (name, value) in options.items():
                _opt = _get_option(section, name, warn)
                if _opt is not None:
                    _opt.validator(value)

    def write_config(self):
        with open(CFG_FILE, "w") as config_file:
            self.vconf.write(config_file)

    def set_config(self, section, name, value, write=True):
        """
        Set a configuration value and (optionally) save to config file

        Parameters
        ----------
        section : str
            The section name for the config option

        name : str
            The name of the config value to set in under the section

        value : any
            The value to be associated with the name

        write : bool, optional(default=True)
            Bool indicating if the config file should be written to disk after
            updating the value
        """
        _validate_config_setting(section, name, value)
        self.vconf.set(section, name, value)

        if write:
            self.write_config()

    def _get_sec_opt(self, key):
        parts = key.split(".")
        if len(parts) != 2:
            raise ValueError("key must have form SECTION.option")

        return parts

    def __setitem__(self, key, val):
        section, name = self._get_sec_opt(key)
        if val is not None:
            self.set_config(section, name, val)

    def __getitem__(self, key, default=None):
        section, name = self._get_sec_opt(key)
        if self.vconf.has_option(section, name):
            return self.vconf.get(section, name)

        if default is not None:
            return default

        return _get_option(section, name).default


options = _DictOptions()


def describe_options():
    msg = "qeds configuration options are:\n\n"
    for (sec, options) in _valid_options.items():
        msg += sec + "\n"
        for o in options:
            msg += str(o) + "\n"

        msg += "\n"

    print(msg)


# --------------------- #
# logging configuration #
# --------------------- #

def setup_logger(module):
    log = logging.getLogger(module)
    log.setLevel(options["options.log_level"])
    return log



import os

import pandas as pd
import requests
import warnings

from ..config import options, setup_logger

from .util import SOCRATA_STATUS_CODE_REASONS

LOGGER = setup_logger(__name__)


class SocrataData(object):

    _socrata_datasources = {
            "NYCOpenData": "https://data.cityofnewyork.us"
            }

    def __init__(self, dataset, datasource, key=None):
        """
        Socrata Open Data: https://dev.socrata.com/

        Parameters
        ----------
        dataset : string
            TODO

        datasource : string
            TODO

        key : string, optional
            The API registrationKey to be used when requesting data.
            There are three possibilities for determining which api key is
            used:

            1. User argument: if the argument supplied here is a valid
               Socrata app token, then that value will be used
            2. Environment variable: if step 1 fails, we will attempt to use
               the value of the environment variable {key_env_name}.
            3. qeds conf: if that fails, we will attempt to look up the
               appropriate value in the qeds configuration file

            If all three of those fail then we will revert to making public
            requests which may be throttled and thus be slower.

            If either step 1 or step 2 succeeds, we will store the api key
            in the qedsm conf file under `socrata.api_key`. This means you
            should only need to supply a key once per machine.
        """.format(key_env_name=options["socrata.environment_variable"])
        # Check whether it is a valid datasource
        if datasource not in self._socrata_datasources.keys():
            msg = "Dataset not found in list of Socrata datasets"
            msg += "\n\nIt either (1) does not exist, "
            msg += "(2) is not implemented, or (3) you have made a typo"
            raise ValueError(msg)

        # Check whether it is valid dataset
        if len(dataset) != 9:
            raise ValueError("Dataset must be of form XXXX-ZZZZ")

        # Find key
        update_config = True
        if key is None:
            KEY_ENV_NAME = options["socrata.environment_variable"]
            if KEY_ENV_NAME in os.environ:
                key = os.environ[KEY_ENV_NAME]
            elif options["socrata.api_key"] is not None:
                key = options["socrata.api_key"]
                update_config = False
            else:
                url = "https://dev.socrata.com/register"
                msg = "\nSocrata API key not detected"
                msg += " please obtain one from {}".format(url)
                msg += " and call `qeds.options['socrata.api_key']=key`"
                msg += "\nFor now we will use no key. Socrata might limit"
                msg += " our usage."
                warnings.warn(msg)

        # If we don't already have they key then save it
        if update_config:
            LOGGER.debug("Saving socrata API key")
            options["socrata.api_key"] = key

        # Get information ready for requests
        self.dataset = dataset
        self.datasource = datasource
        self.key = key
        self.base_url = self._socrata_datasources[datasource]
        self.headers = {"X-APP-TOKEN": key} if key is not None else {}

    def get(self, SoQL={}, limit=None):
        """
        Get self.dataset from self.datasource and convert to a
        pandas DataFrame

        Parameters
        ----------
        SoQL: Dict
            A dictionary which contains the filters and other restrictions
            that you'd like to place on the data as it is brought into
            the DataFrame https://dev.socrata.com/docs/queries/

        Returns
        -------
        df : pandas.DataFrame
            A pandas DataFrame contianing the requested series

        """
        # Build request that we will make
        req_url = self.base_url + "/resource/{}.json".format(self.dataset)

        # If we don't give it a limit, try and select all rows
        if limit is None:
            # Count how many rows there are using the `count(*)` option
            lim_params = SoQL.copy()
            lim_params.update({"$select": "count(*)"})

            # Make request and figure out limit
            LOGGER.debug("Requesting data from {} with params {}".format(
                req_url, lim_params
            ))
            req_count = requests.get(
                req_url, headers=self.headers, params=lim_params
            )
            limit = int(req_count.json()[0]["count"])

        # Add limit to params and request data
        SoQL["$limit"] = limit
        mystr = "Requesting data from {} with params {}"
        LOGGER.debug(mystr.format(req_url, SoQL))
        records = requests.get(req_url, headers=self.headers, params=SoQL)

        # Raise error if not success
        if records.status_code not in [200]:
            raise ValueError(SOCRATA_STATUS_CODE_REASONS[records.status_code])

        # Construct data
        df = pd.DataFrame.from_records(records.json())

        return df



KEY_ENV_NAME = "SOCRATA_API_KEY"

# see https://dev.socrata.com/docs/response-codes.html
SOCRATA_STATUS_CODE_REASONS = {
    200: "Success",
    202: "Still processing",
    400: "Your request did not follow the correct syntax.",
    401: "You are not authorized to make this request.",
    403: "You are not authorized to access this resource",
    404: "Your request was not found and/or does not exist.",
    429: "You have made too many requests",
    500: "Server error"
}



from . import config
from . import shopify
from . import loader
from . import retrievers

from .config import options
from .loader import load, retrieve, available

from .bls import *  # noqa: F403,F401
from .socrata import *  # noqa: F403,F401
from .uscensus import *  # noqa: F403,F401


__all__ = [
    "config", "shopify", "loader", "retrievers", "options", "load",
    "retrieve", "available"
]



"""
Adapted from https://github.com/pypa/sampleproject/blob/master/setup.py
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import re
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

descr = "Python library for loading/cleaning data used in qeds training"

# get the version from the version.py file
with open(path.join(here, "qeds", "version.py"), encoding="utf-8") as f:
    contents = f.read()
    version = re.match(r'__version__ = "(.+)"', contents).group(1)

setup(
    name="qeds",

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description=descr,
    long_description=long_description,

    # The project"s main homepage.
    # url="https://github.com/sglyon/uscensus",

    # Author details
    author="qeds Data",
    author_email="admin@qedsdata.com",

    # Choose your license
    license="BSD-3",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",

        # Indicate who your project is intended for
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Database :: Front-Ends",
        "Topic :: Education",

        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: BSD License",

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6"
    ],

    # What does your project relate to?
    keywords="data pandas economics census education",

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests"]),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip"s
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["pandas", "requests"],
)



"""
This file provides the `data_read` function which reads data
from a particular folder on the computer
"""
import json
import os
import pandas as pd
from .config import options, setup_logger
from .util import _ensure_dir

LOGGER = setup_logger(__name__)

_ensure_dir(options["PATHS.data"])

_METADATA_FN = os.path.join(options["PATHS.data"], "metadata.json")
if not os.path.isfile(_METADATA_FN):
    with open(_METADATA_FN, "w") as f:
        json.dump(dict(), f)


def _update_metadata(name, metadata):
    with open(_METADATA_FN, "r") as f:
        current = json.load(f)

    current[name] = metadata
    with open(_METADATA_FN, "w") as f:
        json.dump(current, f)

    return current


def _get_metadata(name):
    with open(_METADATA_FN, "r") as f:
        current = json.load(f)

    return current.get(name, dict())


def load(name, kwargs={}):
    # Create the file name that corresponds to where this file
    # should be stored
    EXTENSION = options["options.file_format"]
    fn = os.path.join(options["PATHS.data"], name) + "." + EXTENSION

    def _update_using_meta(df):
        meta = _get_metadata(name)
        if len(meta.get("index", dict())) > 0:
            if EXTENSION in ["csv", "feather"]:
                df.set_index(meta["index"], inplace=True)

        return df

    # Check whether the file exists
    if os.path.exists(fn):
        LOGGER.debug("Loading data from {}".format(fn))
        # If it exists, read it in directly
        if EXTENSION == "csv":
            return _update_using_meta(pd.read_csv(fn, **kwargs))
        elif EXTENSION == "pkl":
            return _update_using_meta(pd.read_pickle(fn, **kwargs))
        elif EXTENSION == "feather":
            return _update_using_meta(pd.read_feather(fn, **kwargs))

        else:
            raise ValueError("Unknown extension type {}".format(EXTENSION))
    else:
        return retrieve(name)


def retrieve(name, kwargs={}):
    """
    Retrieves a dataset according to the instructions maintained in
    another function `_data_retrieve_{name}` and saves it to your
    computer

    Parameters
    ----------
    name : string
        The name of the dataset that you want to retrieve from online. It
        is saved at "$qeds/data/name.{file_ending}"

    Returns
    -------
    None

    """
    from .retrievers import __dict__ as retrievers

    func_name = "_retrieve_{}".format(name)
    func = retrievers.get(func_name, None)
    if func is None:
        msg = "The dataset name that you gave ({}) is not on your computer \n"
        msg += "and can not be retrieved by our library. Are you sure \n"
        msg += "you typed it correctly?"
        raise ValueError(msg.format(name))

    # Call retrieval function
    df, metadata = func()
    _update_metadata(name, metadata)

    # Save file
    EXTENSION = options["options.file_format"]
    fn = os.path.join(options["PATHS.data"], name + "." + EXTENSION)

    # Check whether the folder exists and if not create it
    if not os.path.exists(options["PATHS.data"]):
        os.makedirs(options["PATHS.data"])

    # Save data into folder
    if EXTENSION == "csv":
        df.to_csv(fn, **kwargs)
    elif EXTENSION == "pkl":
        df.to_pickle(fn, **kwargs)
    elif EXTENSION == "feather":
        df.reset_index().to_feather(fn, **kwargs)

    return df


def available(name=None):
    """
    Return a list of available datasets

    Parameters
    ----------
    name : string (optional)
        A string used to filter datasets. Datasets that include ``name``
        in the name of the dataset are kept. If nothing is given for this
        argument, no filtering is done

    Returns
    -------
    datasets : list(string)
        A list of available
    """

    from .retrievers import __dict__ as retrievers

    skip = len("_retrieve_")
    out = (k[skip:] for k in retrievers.keys() if k.startswith("_retrieve_"))

    if name is None:
        return list(out)

    return list(k for k in out if name in k)
